/**
 * qdb.c
 **/
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <memory.h>
#include <math.h>

#include "qdb.h"

int		qdbError = QDBERR_INIT;
qdb_t	*qdb = NULL;
char	*_currkey = NULL;
char	*_currvalue = NULL;

/**
 * hash function used to generate
 * a index that spreads a key among
 * the available buckets
 **/
int		_QDB_Hash (char *input)
{
	int		result = 0;
	int		i;
	char	*p = input;
	
	while (*p)
	{
		i = (int) *p;
		i++;
		i = i<<2;
		i ^= (int) *p++;
		result += i;
	}
	
	result &= (QDB_BUCKETS - 1);	
	return	result;
}

int		QDB_Init (uint szkey, uint szvalue) {
	int	i;
	
	if (qdb)
	{
		QDB_Down ();
	}
	
	qdb = (qdb_t *)malloc (sizeof (qdb_t));
	
	qdb->sz_key = sizeof (char) * szkey;
	qdb->keys = (char *)malloc (qdb->sz_key);			
	qdb->sz_value = sizeof (char) * szvalue;	
	qdb->values = (char *)malloc (qdb->sz_value);		
	qdb->buckets = (qdblink_t *)malloc (sizeof (qdblink_t) * QDB_BUCKETS);
	for (i = 0; i < QDB_BUCKETS; i++)
	{
		qdb->buckets[i].pair = (pair_t *)malloc (sizeof (pair_t));
		qdb->buckets[i].next = NULL;
	}

	QDB_Clear ();
	return	QDBERR_OK;
}

int		QDB_Down () {
	if (qdb)
	{
		QDB_Clear ();
		if (qdb->keys)
		{
			free (qdb->keys);
		}
		
		if (qdb->values)
		{
			free (qdb->values);
		}
		
		free (qdb->buckets);
	}
	
	return	QDBERR_OK;
}

int		QDB_Error () {
	return qdbError;
}

int		QDB_Open (char *name) {
	FILE 	*fin;
	// FIXME
	char	key[256], value[8192], *ptr1, *ptr2, c;
	int		ctrl;
	
	if (qdb)
	{
		QDB_Clear ();
		if ((fin = fopen (name, "rt")) == NULL)
		{
			return QDBERR_FILEIO;
		}
		
		ctrl = 0;
		ptr1 = key;
		ptr2 = value;
		memset (key, '\0', 256);
		memset (value, '\0', 8192);
		while (!feof (fin))
		{
			fread (&c, sizeof(char), 1, fin);
			if ((c == '"') && (ctrl == 0))
			{
				ctrl++;
			}
			else if ((c == '"') && (ctrl == 1))
			{
				ctrl++;
			}
			else if ((c == '=') && (ctrl == 2))
			{
				ctrl++;
			}
			else if ((c == '"') && (ctrl == 3))
			{
				ctrl++;
			}
			else if ((c == '"') && (ctrl == 4))
			{
				ctrl++;
			}			
			else if ((c == ';') && (ctrl == 5))
			{
				ctrl = 0;
				QDB_Put (key, value);
				ptr1 = key;
				ptr2 = value;
				memset (key, '\0', 256);
				memset (value, '\0', 8192);				
			}
			else
			{
				if (c == '\\')
				{
					if ((ctrl == 1) || (ctrl == 4) && (!feof(fin)))
					{
						fread (&c, sizeof(char), 1, fin);
						if (c == 'n')
						{
							c = '\n';
						}
						// Quake control chars
						else if (c == '1')
						{
							c = '\001';
						}
						else if (c == '2')
						{
							c = '\002';
						}
					}
				}
								
				if ((ctrl == 1) && ((ptr1 - key) < 256))
				{
					*ptr1=c;
					ptr1++;
				}
				else if ((ctrl == 4) && ((ptr2 - value) < 8192))
				{
					*ptr2=c;
					ptr2++;
				}					
			}
		}
		
		fclose (fin);
		return	QDBERR_OK;
	}
	
	return	QDBERR_INIT;
}

int		QDB_Clear () {
	int		i;
	qdblink_t	*ptr, *tmp;
	
	if (qdb)
	{
		memset (qdb->keys, '\0', qdb->sz_key);
		qdb->ofs_key = 0;
		memset (qdb->values, '\0', qdb->sz_value);
		qdb->ofs_value = 0;
		
		for (i = 0; i < QDB_BUCKETS; i++)
		{
			ptr = (qdblink_t *)&qdb->buckets[i];
			ptr->pair->key = QDB_NULL;
			ptr->pair->value = QDB_NULL;
			ptr = (qdblink_t *)ptr->next;
			qdb->buckets[i].next = NULL;
			while (ptr)
			{
				tmp = (qdblink_t *)ptr->next;
				if (ptr != (qdblink_t *)&qdb->buckets[i])
				{
					free (ptr);
				}
				
				ptr = tmp;
				tmp = NULL;
			}			
		}	

		qdb->changed = 0;
	}
	
	return	QDBERR_OK;
}

int		QDB_Flush (char *name) {
	int		i;
	qdblink_t	*ptr;
	char	*key, *value;
	FILE	*fout;
	
	if ((qdb) && (qdb->changed))
	{
		if ((fout = fopen (name, "wt")) == NULL)
		{
			return QDBERR_FILEIO;
		}
		
		for (i = 0; i < QDB_BUCKETS; i++)
		{
			ptr = (qdblink_t *) &qdb->buckets[i];
			while ((ptr) && (ptr->pair->key != QDB_NULL) && (ptr->pair->value != QDB_NULL))
			{
				key = &qdb->keys[ptr->pair->key];
				value = &qdb->values[ptr->pair->value];
				fprintf (fout, "\"%s\"=\"%s\";\n", key, value);
				ptr = (qdblink_t *)ptr->next;
			}
		}
		
		qdb->changed = 0;
		fclose (fout);
	}
	
	return	QDBERR_OK;
}

int		QDB_Close (char *name) {
	QDB_Flush (name);
	QDB_Clear ();
	return	QDBERR_OK;
}

int		QDB_HasKey (char *key) {
	int		result = -1;
	int		i, l;
	qdblink_t	*ptr;
	char	*str;
	
	if ((qdb) && (key))
	{
		qdbError = QDBERR_NOTFOUND;
		_currkey = NULL;
		_currvalue = NULL;
		l = strlen (key);
		i = _QDB_Hash (key);
		ptr = (qdblink_t *) &qdb->buckets[i];
		while ((ptr) && (ptr->pair->key != QDB_NULL))
		{
			str = (char *) &qdb->keys[ptr->pair->key];
			if (!strncmp (str, key, l))
			{
				_currkey = str;
				_currvalue = (char *)&qdb->values[ptr->pair->value];
				qdbError = QDBERR_OK;
				return 1;
			}
			
			ptr = (qdblink_t *) ptr->next;
		}
	}
	
	return	result;
}

char	*QDB_Get (char *key) {
	if (qdb)
	{
		QDB_HasKey (key);
		return strdup (_currvalue);
	}
	
	return	NULL;
}

float	QDB_GetF (char *key) {
	float result = 0.0f;
		
	if (qdb)
	{
		if (QDB_HasKey (key) != -1)
		{
			result = atof (_currvalue);
			
		}
	}	
	
	return result;
}

int		QDB_Put (char *key, char *value) {
	int		i, j, k, l;
	qdblink_t	*ptr;
	char	*str, *str2;
	
	if ((qdb) && (key) && (value))
	{
		if (QDB_HasKey (key) != -1)
		{
			i = strlen (value);
			j = strlen (_currvalue);
			// clears the previous value
			memset (_currvalue, '\0', j);
			if (j >= i)
			{				
				// it's safe (and faster) to just copy over the previous value
				memcpy (_currvalue, value, i);
			}
			else
			{
				if ((unsigned int)i < (qdb->sz_value - qdb->ofs_value))
				{
					str = (char *) &qdb->values[qdb->ofs_value];
					memcpy (str, value, i);
					k = qdb->ofs_value;
					l = strlen (key);
					qdb->ofs_value = qdb->ofs_value + i + 1;
					i = _QDB_Hash (key);
					ptr = (qdblink_t *) &qdb->buckets[i];
					while ((ptr) && (ptr->pair->key != QDB_NULL))
					{
						str2 = (char *) &qdb->keys[ptr->pair->key];
						if (!strncmp (str2, key, l))
						{
							// update the value
							ptr->pair->value = k;
							_currkey = str2;
							_currvalue = (char *)&qdb->values[ptr->pair->value];
							qdbError = QDBERR_OK;
							return QDBERR_OK;
						}
						
						ptr = (qdblink_t *) ptr->next;
					}
				}
				else
				{
					// no room for the new value, abort
					return QDBERR_OUTOFMEM;
				}
			}
		}
		else
		{
			i = strlen (key);
			j = strlen (value);
			
			if ((unsigned int)i >= (qdb->sz_key - qdb->ofs_key))
			{
				return QDBERR_OUTOFMEM;
			}
			
			if ((unsigned int)j >= (qdb->sz_value - qdb->ofs_value))
			{
				return QDBERR_OUTOFMEM;
			}				
			
			str = &qdb->keys[qdb->ofs_key];
			str2 = &qdb->values[qdb->ofs_value];
			memcpy (str, key, i);
			memcpy (str2, value, j);
			k = qdb->ofs_key;
			l = qdb->ofs_value;
			qdb->ofs_key = qdb->ofs_key + i + 1;
			qdb->ofs_value = qdb->ofs_value + j + 1;
			
			i = _QDB_Hash (key);
			ptr = (qdblink_t *) &qdb->buckets[i];
			while ((ptr) && (ptr->pair->key != QDB_NULL) && (ptr->next != NULL))
			{
				ptr = (qdblink_t *)ptr->next;
			}
			
			if (ptr)
			{
				if (ptr->pair->key == QDB_NULL)
				{
					ptr->pair->key = k;
					ptr->pair->value = l;
				}
				else if (ptr->next == NULL)
				{
					ptr->next = (qdblink_t *) malloc (sizeof (qdblink_t));
					ptr = (qdblink_t *) ptr->next;
					ptr->pair = (pair_t *) malloc (sizeof (pair_t));
					ptr->next = NULL;
					ptr->pair->key = k;
					ptr->pair->value = l;					
				}
			}
		}
		
		qdb->changed = 1;
	}
	else if (!qdb)
	{
		return QDBERR_INIT;
	}
	else if (!key)
	{
		return QDBERR_BADKEY;
	}
	else 
	{
		return QDBERR_BADVALUE;
	}
	
	return	QDBERR_OK;
}

int		QDB_PutF (char *key, float value) {
	char	buf[32];
	int		i = (int) value;
	
	if ((qdb) && (key))
	{
		if (fabs(value) - abs(i) >= 0.006f)
		{
			sprintf (buf, "%6.2f", value);
		}
		else
		{
			sprintf (buf, "%i", i);
		}
		
		return QDB_Put (key, buf);
	}
	
	return	QDBERR_INIT;
}

int		QDB_Remove (char *key) {
	int		i, l;
	qdblink_t	*ptr, *prev;
	char	*str;

	if ((qdb) && (key))
	{
		l = strlen (key);
		i = _QDB_Hash (key);
		ptr = (qdblink_t *) &qdb->buckets[i];
		prev = ptr;
		while ((ptr) && (ptr->pair->key != QDB_NULL))
		{
			str = (char *) &qdb->keys[ptr->pair->key];
			if (!strncmp (str, key, l))
			{
				memset (str, '\0', l);
				str = (char *) &qdb->values[ptr->pair->value];
				l = strlen (str);
				memset (str, '\0', l);				
				prev->next = (qdblink_t *)ptr->next;
				free (ptr->pair);
				free (ptr);
				ptr = NULL;
				return QDBERR_OK;
			}
			
			prev = ptr;
			ptr = (qdblink_t *) ptr->next;
		}
		
		return QDBERR_NOTFOUND;
	}
	
	return	QDBERR_OK;
}
